from enum import Enum
from ibm_watsonx_data_integration.services.datastage.models.connections.base_connection import BaseConnection
from ibm_watsonx_data_integration.services.datastage.models.enums import COGNOS_ANALYTICS_CONNECTION
from pydantic import ConfigDict, Field
from typing import ClassVar


class AuthMethod(Enum):
    anonymous = "anonymous"
    username_apikey_namespace_cpdca = "username_apikey_namespace_cpdca"
    username_password_namespace_cpdca = "username_password_namespace_cpdca"
    cognos_analytics_apikey = "cognos_analytics_apikey"
    username_password_namespace = "username_password_namespace"


class CognosAnalyticsConn(BaseConnection):
    model_config = ConfigDict(populate_by_name=True)

    datasource_type: ClassVar[str] = "11f3029d-a1cf-4c4d-b8e7-64422fa54a94"
    asset_id: str | None = None
    proj_id: str | None = None
    name: str = Field(None, alias="name")
    ds_host: str | None = Field(None, alias="_host")
    ds_physical_collection: str | None = Field(None, alias="_physical_collection")
    ds_port: str | None = Field(None, alias="_port")
    access_token: str = Field(None, alias="access_token")
    api_key: str = Field(None, alias="api_key")
    authentication_method: COGNOS_ANALYTICS_CONNECTION.AuthMethod | None = Field(None, alias="auth_method")
    cluster_access_token: str | None = Field(None, alias="cluster_access_token")
    cluster_user_name: str | None = Field(None, alias="cluster_user_name")
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    use_my_platform_login_credentials: bool | None = Field(False, alias="inherit_access_token")
    namespace_id: str = Field(None, alias="namespace_id")
    password: str = Field(None, alias="password")
    secure_gateway_id: str | None = Field(None, alias="sg_gateway_id")
    sg_host_original: str | None = Field(None, alias="sg_host_original")
    secure_gateway_as_http_proxy: bool | None = Field(None, alias="sg_http_proxy")
    secure_gateway_security_token: str | None = Field(None, alias="sg_security_token")
    secure_gateway_service_url: str | None = Field(None, alias="sg_service_url")
    satellite_client_certificate: str | None = Field(None, alias="sl_client_cert")
    satellite_client_private_key: str | None = Field(None, alias="sl_client_private_key")
    satellite_connector_id: str | None = Field(None, alias="sl_connector_id")
    satellite_endpoint_host: str | None = Field(None, alias="sl_endpoint_host")
    satellite_endpoint_display_name: str | None = Field(None, alias="sl_endpoint_name")
    satellite_endpoint_port: int | None = Field(None, alias="sl_endpoint_port")
    sl_host_original: str | None = Field(None, alias="sl_host_original")
    satellite_as_http_proxy: bool | None = Field(None, alias="sl_http_proxy")
    satellite_location_id: str | None = Field(None, alias="sl_location_id")
    satellite_service_url: str | None = Field(None, alias="sl_service_url")
    ssl_certificate: str | None = Field(None, alias="ssl_certificate")
    url: str = Field(None, alias="url")
    use_anonymous_access: bool | None = Field(False, alias="use_anonymous_access")
    username: str = Field(None, alias="username")
    vaulted_properties: str | None = Field(None, alias="vaulted_properties")
    ssl_certificate_file: str | None = Field(None, alias="ssl_certificate_file")
    hidden_dummy_property1: str | None = Field(None, alias="hiddenDummyProperty1")
    hidden_dummy_property2: str | None = Field(None, alias="hiddenDummyProperty2")

    def validate(self):
        include = set()
        exclude = set()

        (
            include.add("access_token")
            if (
                ((not self.defer_credentials))
                and (not self.use_anonymous_access)
                and (not self.username)
                and (not self.namespace_id)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value != "anonymous"
                        )
                        or (self.authentication_method != "anonymous")
                    )
                )
                and (self.use_my_platform_login_credentials)
            )
            else exclude.add("access_token")
        )
        (
            include.add("password")
            if (
                ((not self.defer_credentials))
                and (not self.access_token)
                and (not self.use_my_platform_login_credentials)
                and (not self.use_anonymous_access)
                and (
                    (
                        (
                            self.authentication_method
                            and (
                                (
                                    hasattr(self.authentication_method, "value")
                                    and self.authentication_method.value == "username_password_namespace"
                                )
                                or (self.authentication_method == "username_password_namespace")
                            )
                        )
                    )
                    or (
                        self.authentication_method
                        and (
                            (
                                hasattr(self.authentication_method, "value")
                                and self.authentication_method.value == "username_password_namespace_cpdca"
                            )
                            or (self.authentication_method == "username_password_namespace_cpdca")
                        )
                    )
                )
            )
            else exclude.add("password")
        )
        include.add("authentication_method") if (not self.defer_credentials) else exclude.add("authentication_method")
        (
            include.add("namespace_id")
            if (
                ((not self.defer_credentials))
                and (not self.access_token)
                and (not self.use_my_platform_login_credentials)
                and (
                    (
                        (
                            self.authentication_method
                            and (
                                (
                                    hasattr(self.authentication_method, "value")
                                    and self.authentication_method.value == "username_apikey_namespace_cpdca"
                                )
                                or (self.authentication_method == "username_apikey_namespace_cpdca")
                            )
                        )
                    )
                    or (
                        self.authentication_method
                        and (
                            (
                                hasattr(self.authentication_method, "value")
                                and self.authentication_method.value == "username_password_namespace"
                            )
                            or (self.authentication_method == "username_password_namespace")
                        )
                    )
                    or (
                        self.authentication_method
                        and (
                            (
                                hasattr(self.authentication_method, "value")
                                and self.authentication_method.value == "username_password_namespace_cpdca"
                            )
                            or (self.authentication_method == "username_password_namespace_cpdca")
                        )
                    )
                )
            )
            else exclude.add("namespace_id")
        )
        (
            include.add("satellite_connector_id")
            if (((not self.secure_gateway_id)) and (not self.satellite_location_id))
            else exclude.add("satellite_connector_id")
        )
        (
            include.add("api_key")
            if (
                ((not self.defer_credentials))
                and (not self.access_token)
                and (not self.use_my_platform_login_credentials)
                and (not self.use_anonymous_access)
                and (
                    (
                        (
                            self.authentication_method
                            and (
                                (
                                    hasattr(self.authentication_method, "value")
                                    and self.authentication_method.value == "cognos_analytics_apikey"
                                )
                                or (self.authentication_method == "cognos_analytics_apikey")
                            )
                        )
                    )
                    or (
                        self.authentication_method
                        and (
                            (
                                hasattr(self.authentication_method, "value")
                                and self.authentication_method.value == "username_apikey_namespace_cpdca"
                            )
                            or (self.authentication_method == "username_apikey_namespace_cpdca")
                        )
                    )
                )
            )
            else exclude.add("api_key")
        )
        (
            include.add("use_my_platform_login_credentials")
            if (
                ((not self.defer_credentials))
                and (not self.use_anonymous_access)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "username_password_namespace"
                        )
                        or (self.authentication_method == "username_password_namespace")
                    )
                )
            )
            else exclude.add("use_my_platform_login_credentials")
        )
        (
            include.add("satellite_location_id")
            if (((not self.secure_gateway_id)) and (not self.satellite_connector_id))
            else exclude.add("satellite_location_id")
        )
        (
            include.add("username")
            if (
                ((not self.defer_credentials))
                and (not self.access_token)
                and (not self.use_my_platform_login_credentials)
                and (not self.use_anonymous_access)
                and (
                    (
                        (
                            self.authentication_method
                            and (
                                (
                                    hasattr(self.authentication_method, "value")
                                    and self.authentication_method.value == "username_apikey_namespace_cpdca"
                                )
                                or (self.authentication_method == "username_apikey_namespace_cpdca")
                            )
                        )
                    )
                    or (
                        self.authentication_method
                        and (
                            (
                                hasattr(self.authentication_method, "value")
                                and self.authentication_method.value == "username_password_namespace"
                            )
                            or (self.authentication_method == "username_password_namespace")
                        )
                    )
                    or (
                        self.authentication_method
                        and (
                            (
                                hasattr(self.authentication_method, "value")
                                and self.authentication_method.value == "username_password_namespace_cpdca"
                            )
                            or (self.authentication_method == "username_password_namespace_cpdca")
                        )
                    )
                )
            )
            else exclude.add("username")
        )
        (
            include.add("use_anonymous_access")
            if (
                ((not self.defer_credentials))
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "anonymous"
                        )
                        or (self.authentication_method == "anonymous")
                    )
                )
            )
            else exclude.add("use_anonymous_access")
        )
        (
            include.add("hidden_dummy_property2")
            if (self.hidden_dummy_property1)
            else exclude.add("hidden_dummy_property2")
        )
        (
            include.add("hidden_dummy_property1")
            if (self.hidden_dummy_property2)
            else exclude.add("hidden_dummy_property1")
        )
        include.add("access_token") if (self.hidden_dummy_property1) else exclude.add("access_token")
        include.add("vaulted_properties") if (self.hidden_dummy_property1) else exclude.add("vaulted_properties")
        (
            include.add("use_my_platform_login_credentials")
            if (self.hidden_dummy_property1)
            else exclude.add("use_my_platform_login_credentials")
        )
        include.add("ssl_certificate_file") if (self.hidden_dummy_property1) else exclude.add("ssl_certificate_file")

        (
            include.add("satellite_connector_id")
            if (not self.secure_gateway_id) and (not self.satellite_location_id)
            else exclude.add("satellite_connector_id")
        )
        (
            include.add("api_key")
            if (not self.defer_credentials)
            and (not self.access_token)
            and (self.use_my_platform_login_credentials != "yes")
            and (not self.use_anonymous_access)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "cognos_analytics_apikey" in str(self.authentication_method.value)
                    )
                    or ("cognos_analytics_apikey" in str(self.authentication_method))
                )
                and self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "username_apikey_namespace_cpdca" in str(self.authentication_method.value)
                    )
                    or ("username_apikey_namespace_cpdca" in str(self.authentication_method))
                )
            )
            else exclude.add("api_key")
        )
        include.add("authentication_method") if (not self.defer_credentials) else exclude.add("authentication_method")
        (
            include.add("use_my_platform_login_credentials")
            if (not self.defer_credentials)
            and (not self.use_anonymous_access)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value == "username_password_namespace"
                    )
                    or (self.authentication_method == "username_password_namespace")
                )
            )
            else exclude.add("use_my_platform_login_credentials")
        )
        (
            include.add("use_anonymous_access")
            if (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "anonymous")
                    or (self.authentication_method == "anonymous")
                )
            )
            else exclude.add("use_anonymous_access")
        )
        (
            include.add("access_token")
            if (not self.defer_credentials)
            and (not self.use_anonymous_access)
            and (not self.username)
            and (not self.namespace_id)
            and (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value != "anonymous")
                    or (self.authentication_method != "anonymous")
                )
            )
            and (self.use_my_platform_login_credentials == "yes")
            else exclude.add("access_token")
        )
        (
            include.add("password")
            if (not self.defer_credentials)
            and (not self.access_token)
            and (self.use_my_platform_login_credentials != "yes")
            and (not self.use_anonymous_access)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "username_password_namespace" in str(self.authentication_method.value)
                    )
                    or ("username_password_namespace" in str(self.authentication_method))
                )
                and self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "username_password_namespace_cpdca" in str(self.authentication_method.value)
                    )
                    or ("username_password_namespace_cpdca" in str(self.authentication_method))
                )
            )
            else exclude.add("password")
        )
        (
            include.add("username")
            if (not self.defer_credentials)
            and (not self.access_token)
            and (self.use_my_platform_login_credentials != "yes")
            and (not self.use_anonymous_access)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "username_apikey_namespace_cpdca" in str(self.authentication_method.value)
                    )
                    or ("username_apikey_namespace_cpdca" in str(self.authentication_method))
                )
                and self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "username_password_namespace" in str(self.authentication_method.value)
                    )
                    or ("username_password_namespace" in str(self.authentication_method))
                )
                and self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "username_password_namespace_cpdca" in str(self.authentication_method.value)
                    )
                    or ("username_password_namespace_cpdca" in str(self.authentication_method))
                )
            )
            else exclude.add("username")
        )
        (
            include.add("namespace_id")
            if (not self.defer_credentials)
            and (not self.access_token)
            and (self.use_my_platform_login_credentials != "yes")
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "username_apikey_namespace_cpdca" in str(self.authentication_method.value)
                    )
                    or ("username_apikey_namespace_cpdca" in str(self.authentication_method))
                )
                and self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "username_password_namespace" in str(self.authentication_method.value)
                    )
                    or ("username_password_namespace" in str(self.authentication_method))
                )
                and self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "username_password_namespace_cpdca" in str(self.authentication_method.value)
                    )
                    or ("username_password_namespace_cpdca" in str(self.authentication_method))
                )
            )
            else exclude.add("namespace_id")
        )
        (
            include.add("satellite_location_id")
            if (not self.secure_gateway_id) and (not self.satellite_connector_id)
            else exclude.add("satellite_location_id")
        )
        return include, exclude

    @classmethod
    def from_dict(cls, properties):
        return cls.model_construct(**properties)



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing Flow Model."""

from abc import ABC, abstractmethod
from collections.abc import Callable
from ibm_watsonx_data_integration.common.models import BaseModel
from typing import Any, ClassVar, TypeVar
from typing_extensions import override

F = TypeVar("F", bound="Flow")


class Flow(BaseModel):
    """Represents a generic Flow object."""

    _flow_registry: ClassVar[dict[str, type[F]]] = {}

    @classmethod
    def register(cls: type[F], flow_type: str) -> Callable[[type[F]], type[F]]:
        """Class decorator to register a new Flow subclass.

        Args:
            flow_type (str):
                The key under which the subclass will be stored in `flow_registry`.

        Returns:
            Callable[[Type[F]], Type[F]]:
                A decorator that registers the subclass and returns it unchanged.
        """

        def inner(subclass: type[F]) -> type[F]:
            cls._flow_registry[flow_type] = subclass
            return subclass

        return inner


class PayloadExtender(ABC):
    """Interface class for flows with custom payload logic.

    This interface should be implemented by any flow that requires custom logic for generating a job's payload.
    It ensures a consistent contract for flows that need to override the default payload creation behavior.

    :meta: private
    """

    @abstractmethod
    def extend(self, payload: dict[str, Any], flow: Flow) -> dict[str, Any]:
        """Here we should modify and return payload for job creation."""
        raise NotImplementedError


class DefaultFlowPayloadExtender(PayloadExtender):
    """Default payload extender which setup only `asset_ref`.

    :meta: private
    """

    @override
    def extend(self, payload: dict[str, Any], flow: Flow) -> dict[str, Any]:
        payload["asset_ref"] = flow.flow_id
        return payload



# IBM Confidential
# PID 5900-BAF
# Copyright StreamSets Inc., an IBM Company 2025

"""Contains classes responsible for generating python script preamble."""

import textwrap
from abc import ABC, abstractmethod
from ibm_watsonx_data_integration.common.constants import (
    PROD_BASE_API_URL,
    PROD_BASE_URL,
)
from typing_extensions import override


class Preamble(ABC):
    """Script preamble containing required imports and class initialization."""

    def __init__(
        self,
        source_data: dict,
        base_url: str = PROD_BASE_URL,
        base_api_url: str = PROD_BASE_API_URL,
        api_key_env_var_name: str = "IBM_CLOUD_API_KEY",
    ) -> None:
        """The __init__ of the Preamble class.

        Args:
            source_data: Raw flow data.
            base_url: URL for IBM Cloud.
            base_api_url: URL for API endpoints.
            api_key_env_var_name: Environment variable which contains API Key.
        """
        self._source_data = source_data
        self._base_url = base_url
        self._base_api_url = base_api_url
        self._api_key_var_name = api_key_env_var_name

    @abstractmethod
    def __str__(self) -> str:
        """Here we should return preamble string representation."""


class StreamingPreamble(Preamble):
    """Preamble for streaming flow recreation scripts."""

    @override
    def __str__(self) -> str:
        return textwrap.dedent(f"""\
import os
from ibm_watsonx_data_integration import Platform
from ibm_watsonx_data_integration.common.auth import IAMAuthenticator


auth = IAMAuthenticator(
    api_key=os.getenv("{self._api_key_var_name}"),
    base_auth_url="{self._base_url}",
)
platform = Platform(
    auth=auth,
    base_url="{self._base_url}",
    base_api_url="{self._base_api_url}",
)

project = platform.projects.get(project_id="{self._source_data["metadata"]["project_id"]}")
env = project.environments.get(environment_id="{self._source_data["entity"]["streamsets_flow"]["environment_id"]}")
flow = project.create_flow(name="{self._source_data["metadata"]["name"]}", environment=env, description="{self._source_data["metadata"]["description"]}")
""")  # noqa: E501



from enum import Enum
from ibm_watsonx_data_integration.services.datastage.models.connections.base_connection import BaseConnection
from ibm_watsonx_data_integration.services.datastage.models.enums import DROPBOX_CONNECTION
from pydantic import ConfigDict, Field
from typing import ClassVar


class AuthMethod(Enum):
    accesstoken = "accesstoken"
    refreshtoken = "refreshtoken"


class DropboxConn(BaseConnection):
    model_config = ConfigDict(populate_by_name=True)

    datasource_type: ClassVar[str] = "507b850c-f4a1-41d7-ad64-4182a1264014"
    asset_id: str | None = None
    proj_id: str | None = None
    name: str = Field(None, alias="name")
    ds_host: str | None = Field(None, alias="_host")
    ds_port: str | None = Field(None, alias="_port")
    access_token: str = Field(None, alias="access_token")
    app_key_client_id: str = Field(None, alias="app_key")
    app_secret_client_secret: str = Field(None, alias="app_secret")
    authentication_method: DROPBOX_CONNECTION.AuthMethod | None = Field(
        DROPBOX_CONNECTION.AuthMethod.accesstoken, alias="auth_method"
    )
    cluster_access_token: str | None = Field(None, alias="cluster_access_token")
    cluster_user_name: str | None = Field(None, alias="cluster_user_name")
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    refresh_token: str = Field(None, alias="refresh_token")
    vaulted_properties: str | None = Field(None, alias="vaulted_properties")
    hidden_dummy_property1: str | None = Field(None, alias="hiddenDummyProperty1")
    hidden_dummy_property2: str | None = Field(None, alias="hiddenDummyProperty2")

    def validate(self):
        include = set()
        exclude = set()

        (
            include.add("access_token")
            if (
                ((not self.refresh_token))
                and (not self.app_key_client_id)
                and (not self.app_secret_client_secret)
                and (not self.defer_credentials)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "accesstoken"
                        )
                        or (self.authentication_method == "accesstoken")
                    )
                )
            )
            else exclude.add("access_token")
        )
        (
            include.add("refresh_token")
            if (
                ((not self.access_token))
                and (not self.defer_credentials)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "refreshtoken"
                        )
                        or (self.authentication_method == "refreshtoken")
                    )
                )
            )
            else exclude.add("refresh_token")
        )
        (
            include.add("app_key_client_id")
            if (
                ((not self.access_token))
                and (not self.defer_credentials)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "refreshtoken"
                        )
                        or (self.authentication_method == "refreshtoken")
                    )
                )
            )
            else exclude.add("app_key_client_id")
        )
        (
            include.add("app_secret_client_secret")
            if (
                ((not self.access_token))
                and (not self.defer_credentials)
                and (
                    self.authentication_method
                    and (
                        (
                            hasattr(self.authentication_method, "value")
                            and self.authentication_method.value == "refreshtoken"
                        )
                        or (self.authentication_method == "refreshtoken")
                    )
                )
            )
            else exclude.add("app_secret_client_secret")
        )
        (
            include.add("hidden_dummy_property2")
            if (self.hidden_dummy_property1)
            else exclude.add("hidden_dummy_property2")
        )
        (
            include.add("hidden_dummy_property1")
            if (self.hidden_dummy_property2)
            else exclude.add("hidden_dummy_property1")
        )
        include.add("vaulted_properties") if (self.hidden_dummy_property1) else exclude.add("vaulted_properties")

        (
            include.add("app_secret_client_secret")
            if (not self.access_token)
            and (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "refreshtoken" in str(self.authentication_method.value)
                    )
                    or ("refreshtoken" in str(self.authentication_method))
                )
            )
            else exclude.add("app_secret_client_secret")
        )
        (
            include.add("app_key_client_id")
            if (not self.access_token)
            and (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "refreshtoken" in str(self.authentication_method.value)
                    )
                    or ("refreshtoken" in str(self.authentication_method))
                )
            )
            else exclude.add("app_key_client_id")
        )
        (
            include.add("access_token")
            if (not self.refresh_token)
            and (not self.app_key_client_id)
            and (not self.app_secret_client_secret)
            and (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "accesstoken" in str(self.authentication_method.value)
                    )
                    or ("accesstoken" in str(self.authentication_method))
                )
            )
            else exclude.add("access_token")
        )
        (
            include.add("refresh_token")
            if (not self.access_token)
            and (not self.defer_credentials)
            and (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "refreshtoken" in str(self.authentication_method.value)
                    )
                    or ("refreshtoken" in str(self.authentication_method))
                )
            )
            else exclude.add("refresh_token")
        )
        return include, exclude

    @classmethod
    def from_dict(cls, properties):
        return cls.model_construct(**properties)



import re
from datetime import datetime

_LOG_LEVEL_MAP = {"E": "ERROR", "I": "INFO", "D": "DEBUG", "W": "WARN"}


def transform_logs(logs: list[str]):
    i = 1
    for l in logs:
        tf = _transform_line(l)
        if tf == l:
            yield l
        else:
            yield str(i) + " " + _transform_line(l)
            i += 1


def _transform_line(line: str) -> str:
    if not line.startswith("##"):
        return line

    results = re.match(
        r"##([A-za-z]) ([A-Za-z0-9-]+) (\d{4}-\d{2}-\d{2}) (\d{2}:\d{2}:\d{2})(?:\([0-9]+\))? (.*)",
        line,
    )
    if not results:
        return line

    level_code = results[1].upper()
    iis_code = results[2].upper()
    date = results[3]
    timestamp = results[4]
    message = results[5]

    dt = datetime.strptime(date, "%Y-%m-%d")

    return " ".join(
        [
            f"{dt.month}/{dt.day:02}/{dt.year}",
            timestamp,
            _LOG_LEVEL_MAP.get(level_code, "INFO"),
            iis_code,
            message,
        ]
    )



from enum import Enum
from pydantic import BaseModel, ConfigDict, Field, field_serializer


class BeforeAfter(Enum):
    before = "Before-Stage"
    after = "After-Stage"

    @classmethod
    def from_str(cls, value: str) -> "BeforeAfter":
        try:
            return cls[value]
        except KeyError:
            raise ValueError(
                f"Unknown BeforeAfter value: {value}. Valid options are: {', '.join(cls._member_names_)}"
            ) from None


class SqlType(Enum):
    BIGINT = "BIGINT"
    BINARY = "BINARY"
    BIT = "BIT"
    CHAR = "CHAR"
    DATE = "DATE"
    DECIMAL = "DECIMAL"
    DOUBLE = "DOUBLE"
    FLOAT = "FLOAT"
    INTEGER = "INTEGER"
    LONGNVARCHAR = "LONGNVARCHAR"
    LONGVARBINARY = "LONGVARBINARY"
    LONGVARCHAR = "LONGVARCHAR"
    NCHAR = "NCHAR"
    NUMERIC = "NUMERIC"
    NVARCHAR = "NVARCHAR"
    REAL = "REAL"
    SMALLINT = "SMALLINT"
    TIME = "TIME"
    TIMESTAMP = "TIMESTAMP"
    TINYINT = "TINYINT"
    UNKNOWN = "UNKNOWN"
    VARBINARY = "VARBINARY"
    VARCHAR = "VARCHAR"

    @classmethod
    def from_str(cls, value: str) -> "SqlType":
        try:
            return cls[value.upper()]
        except KeyError:
            raise ValueError(f"Unknown SQL type: {value}. Valid options are: {', '.join(cls._member_names_)}") from None


class LoopVariable(BaseModel):
    model_config = ConfigDict(validate_by_alias=False, validate_by_name=True, use_enum_values=True)
    name: str = Field(None, alias="Name")
    sql_type: SqlType = Field(None, alias="SqlType")
    derivation: str = Field(None, alias="Derivation")
    length: int = Field(6, alias="Precision")
    scale: int = Field(0, alias="Scale")
    initial_value: str = Field("", alias="InitialValue")
    extended: bool = Field(False, alias="Extended")
    description: str = Field("", alias="Description")

    @field_serializer("length")
    def serialize_length(self, length: int) -> str:
        return str(length)

    @field_serializer("scale")
    def serialize_scale(self, scale: str) -> str:
        return str(scale)

    @field_serializer("extended")
    def serialize_extended(self, extended: bool) -> str:
        return str(extended).lower()


class StageVariable(BaseModel):
    model_config = ConfigDict(validate_by_alias=False, validate_by_name=True, use_enum_values=True)
    name: str = Field(None, alias="Name")
    sql_type: SqlType = Field(None, alias="SqlType")
    derivation: str = Field(None, alias="Derivation")
    length: int = Field(100, alias="Precision")
    scale: int = Field(None, alias="Scale")
    initial_value: str = Field(None, alias="InitialValue")
    extended: bool = Field(False, alias="Extended")
    description: str = Field(None, alias="Description")

    @field_serializer("length")
    def serialize_length(self, length: int) -> str:
        return str(length)

    @field_serializer("scale")
    def serialize_scale(self, scale: str) -> str:
        return str(scale)

    @field_serializer("extended")
    def serialize_extended(self, extended: bool) -> str:
        return str(extended).lower()


class Trigger(BaseModel):
    model_config = ConfigDict(validate_by_alias=False, validate_by_name=True, use_enum_values=True)
    routine_name: str = Field(None, alias="RoutineName")
    before_after: BeforeAfter = Field(BeforeAfter.before, alias="Location")
    arguments: list[str] = Field([], alias="arguments")

    @field_serializer("arguments")
    def serialize_arguments(self, arguments: list[str]) -> dict[str, str]:
        return {f"Argument{i + 1}": arg for i, arg in enumerate(arguments)}


class Constraint(BaseModel):
    model_config = ConfigDict(validate_by_alias=False, validate_by_name=True, use_enum_values=True)
    output_name: str = Field(None, alias="output_name")
    constraint: str = Field(None, alias="TransformerConstraint")
    otherwise_log: bool = Field(False, alias="Reject")
    abort_after_rows: int = Field(0, alias="RowLimit")


class transformer:
    LoopVariable = LoopVariable
    StageVariable = StageVariable
    Trigger = Trigger
    Constraint = Constraint
    SqlType = SqlType
    BeforeAfter = BeforeAfter



from enum import Enum
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    field_serializer,
    model_serializer,
)
from typing import Any, Literal


class Method(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"


class AuthenticationType(Enum):
    none = "NoAuth"
    api_key = "APIKeyAuth"
    bearer_token = "BearerAuth"
    basic_auth = "BasicAuth"
    digest_auth = "DigestAuth"


class AddTo(Enum):
    request_header = "HEADER"
    query_param = "QUERY_PARAMS"


class BodyType(Enum):
    none = "NONE"
    form_data = "FORM_DATA"
    x_www_form_urlencoded = "X_WWW_FORM_URLENCODED"
    raw = "RAW"
    binary = "BINARY"


class CertificateType(Enum):
    pem_text = "PEM_TEXT"
    pem_file = "PEM_FILE"
    store_file = "STORE_FILE"


class StoreType(Enum):
    PKCS12 = "PKCS12"
    JKS = "JKS"
    JCEKS = "JCEKS"


class ActionOnFailure(Enum):
    abort = "ABORT"
    reject = "REJECT"
    ignore = "IGNORE"


class DataOutputControl(Enum):
    no_output = "NO_OUTPUT"
    output_per_iteration = "PER_ITERATION"
    output_when_request_complete = "FINAL_ITERATION"


class DataType(Enum):
    Integer = "INTEGER"
    Float = "FLOAT"
    Long = "LONG"
    Double = "DOUBLE"
    BigInteger = "BIG_INTEGER"
    BigDecimal = "BIG_DECIMAL"
    String = "STRING"
    Boolean = "BOOLEAN"
    List = "LIST"
    Set = "SET"
    Map = "MAP"
    Unknown = "UNKNOWN"


class Authentication(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    same_config: bool = Field(False, alias="inherit", description="The same configuration as request 0")
    authentication_type: AuthenticationType = Field(AuthenticationType.none, alias="type")
    add_to: AddTo = Field(None, alias="add_to")
    key: str = Field(None, alias="key")
    value: str = Field(None, alias="value")
    token: str = Field(None, alias="token")
    username: str = Field(None, alias="username")
    password: str = Field(None, alias="password")
    use_expression_token: bool = Field(None, alias="expression_bearerToken_on")
    use_expression_key: bool = Field(None, alias="expression_authKey_on")
    use_expression_value: bool = Field(None, alias="expression_authValue_on")
    use_expression_username: bool = Field(None, alias="expression_username_on")
    use_expression_password: bool = Field(None, alias="expression_password_on")

    @field_serializer("key")
    def serialize_key(self, key: str | dict | None) -> dict | None:
        if key is None:
            return key
        if isinstance(key, dict):
            return key
        if self.use_expression_key:
            return {"expression": key}
        return {"expression": f"`{key}`"}

    @field_serializer("value")
    def serialize_value(self, value: str | dict | None) -> dict | None:
        if value is None:
            return value
        if isinstance(value, dict):
            return value
        if self.use_expression_value:
            return {"expression": value}
        return {"expression": f"`{value}`"}

    @field_serializer("token")
    def serialize_token(self, token: str | dict | None) -> dict | None:
        if token is None:
            return token
        if isinstance(token, dict):
            return token
        if self.use_expression_token:
            return {"expression": token}
        return {"expression": f"`{token}`"}

    @field_serializer("username")
    def serialize_username(self, username: str | dict | None) -> dict | None:
        if username is None:
            return username
        if isinstance(username, dict):
            return username
        if self.use_expression_username:
            return {"expression": username}
        return {"expression": f"`{username}`"}

    @field_serializer("password")
    def serialize_password(self, password: str | dict | None) -> dict | None:
        if password is None:
            return password
        if isinstance(password, dict):
            return password
        if self.use_expression_password:
            return {"expression": password}
        return {"expression": f"`{password}`"}


class Parameter(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    key: str = Field("", alias="key")
    default_value: str = Field("", alias="default_value")
    derivation: str = Field("", alias="derivation")
    description: str = Field("", alias="description")

    @field_serializer("derivation")
    def serialize_derivation(self, derivation: str | dict | None) -> dict | None:
        if derivation is None:
            return derivation
        if isinstance(derivation, dict):
            return derivation
        return {"expression": derivation}


class Header(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    key: str = Field("", alias="key")
    default_value: str = Field("", alias="default_value")
    derivation: str = Field("", alias="derivation")
    description: str = Field("", alias="description")

    @field_serializer("derivation")
    def serialize_derivation(self, derivation: str | dict | None) -> dict | None:
        if derivation is None:
            return derivation
        if isinstance(derivation, dict):
            return derivation
        return {"expression": derivation}


class Cookie(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    key: str = Field("", alias="key")
    default_value: str = Field("", alias="default_value")
    derivation: str = Field("", alias="derivation")
    description: str = Field("", alias="description")

    @field_serializer("derivation")
    def serialize_derivation(self, derivation: str | dict | None) -> dict | None:
        if derivation is None:
            return derivation
        if isinstance(derivation, dict):
            return derivation
        return {"expression": derivation}


class FormData(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    key: str = Field("", alias="key")
    default_value: str = Field("", alias="default_value")
    derivation: str = Field("", alias="derivation")
    description: str = Field("", alias="description")
    content_type: str = Field("", alias="content_type")
    type: Literal["TEXT", "FILE"] = Field("TEXT", alias="type")

    @field_serializer("derivation")
    def serialize_derivation(self, derivation: str | dict | None) -> dict | None:
        if derivation is None:
            return derivation
        if isinstance(derivation, dict):
            return derivation
        return {"expression": derivation}


class FormURLEncodedData(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    key: str = Field("", alias="key")
    default_value: str = Field("", alias="default_value")
    derivation: str = Field("", alias="derivation")
    description: str = Field("", alias="description")

    @field_serializer("derivation")
    def serialize_derivation(self, derivation: str | dict | None) -> dict | None:
        if derivation is None:
            return derivation
        if isinstance(derivation, dict):
            return derivation
        return {"expression": derivation}


class Body(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    type: BodyType = Field(BodyType.none, alias="type")
    form_data: list[FormData] = Field([], alias="form_data")
    form_urlencoded_data: list[FormURLEncodedData] = Field([], alias="form_urlencoded_data")
    content_type: str = Field("", alias="content_type")
    encoding_type: str = Field("UTF-8", alias="charset")
    source: Literal["TEXT", "FILE", "DATA"] = Field(None, alias="source")
    file_path: str = Field(None, alias="file_path")
    raw_text: str = Field(None, alias="raw_text")
    binary_data: str = Field(None, alias="binary_data")
    use_expression_file_path: bool = Field(None, alias="expression_filePath_on")
    use_expression_text: bool = Field(None, alias="use_expression_text")
    use_expression_data: bool = Field(None, alias="use_expression_data")

    @field_serializer("source")
    def serialize_source(self, source: str | None) -> str | None:
        if source:
            return source
        type = self.type.value if hasattr(self.type, "value") else self.type
        if type not in ["BINARY", "RAW"]:
            return source
        if self.file_path:
            return "FILE"
        elif self.raw_text:
            return "TEXT"
        elif self.binary_data:
            return "DATA"
        return "FILE"

    @computed_field
    @property
    def data(self) -> dict | list:
        if self.type.value == "RAW":
            if self.source == "FILE":
                if self.file_path is None:
                    return {"expression": ""}
                elif isinstance(self.file_path, dict):
                    return self.file_path
                elif self.use_expression_file_path:
                    return {"expression": f"{self.file_path}"}
                else:
                    return {"expression": f"`{self.file_path}`"}
            elif self.source == "TEXT":
                if self.raw_text is None:
                    return {"expression": ""}
                elif isinstance(self.raw_text, dict):
                    return self.raw_text
                elif self.use_expression_text:
                    return {"expression": self.raw_text}
                else:
                    return {"expression": f"`{self.raw_text}`"}
            else:
                self.source = "TEXT"
                if self.raw_text is None:
                    return {"expression": ""}
                elif isinstance(self.raw_text, dict):
                    return self.raw_text
                elif self.use_expression_text:
                    return {"expression": self.raw_text}
                else:
                    return {"expression": f"`{self.raw_text}`"}
        elif self.type.value == "BINARY":
            if self.source == "DATA":
                if self.binary_data is None:
                    return {"expression": ""}
                elif isinstance(self.binary_data, dict):
                    return self.binary_data
                elif self.use_expression_data:
                    return {"expression": self.binary_data}
                else:
                    return {"expression": f"`{self.binary_data}`"}
            elif self.source == "FILE":
                if self.file_path is None:
                    return {"expression": ""}
                elif isinstance(self.file_path, dict):
                    return self.file_path
                elif self.use_expression_file_path:
                    return {"expression": self.file_path}
                else:
                    return {"expression": f"`{self.file_path}`"}
            else:
                self.source = "FILE"
                if self.file_path is None:
                    return {"expression": ""}
                elif isinstance(self.file_path, dict):
                    return self.file_path
                elif self.use_expression_file_path:
                    return {"expression": self.file_path}
                else:
                    return {"expression": f"`{self.file_path}`"}
        elif self.type.value == "FORM_DATA":
            return self.form_data
        elif self.type.value == "X_WWW_FORM_URLENCODED":
            return self.form_urlencoded_data
        elif self.type.value == "NONE":
            return {"expression": ""}
        else:
            return None

    @computed_field
    @property
    def expression_text_on(self) -> bool:
        return self.use_expression_text or self.use_expression_data


class RequestInfo(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    same_config: bool = Field(False, alias="inherit", description="The same configuration as request 0")
    query_parameters: list[Parameter] = Field([], alias="params")
    custom_headers: list[Header] = Field([], alias="headers")
    additional_headers_on: bool = Field(False, alias="additional_headers_on")
    additional_headers: str | None = Field(None, alias="additional_headers")
    custom_cookies: list[Cookie] = Field([], alias="cookies")
    body: Body = Field(Body(), alias="body")

    @field_serializer("additional_headers")
    def serialize_additional_headers(self, additional_headers: str | dict | None) -> dict | None:
        if additional_headers is None:
            return additional_headers
        if isinstance(additional_headers, dict):
            return additional_headers
        return {"expression": additional_headers}


class Response(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    same_config: bool = Field(False, alias="inherit", description="The same configuration as request 0")
    type: Literal["TEXT", "FILE"] = Field("TEXT", alias="target")
    content_type: str = Field("", alias="content_type")
    encoding_type: str = Field("UTF-8", alias="charset")
    file_path: str = Field(None, alias="file_path")
    use_expression_file_path: bool = Field(None, alias="expression_filePath_on")

    @computed_field
    @property
    def content(self) -> dict:
        if self.type == "TEXT":
            return {
                "target": self.type,
            }
        elif self.type == "FILE":
            if self.use_expression_file_path:
                return {
                    "target": self.type,
                    "file_path": {"expression": f"`{self.file_path}`" or ""},
                }
            return {
                "target": self.type,
                "file_path": {"expression": f"{self.file_path}" or ""},
            }


class ServerCertificate(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    certificate_type: CertificateType = Field(CertificateType.pem_text, alias="type")
    certificate: str = Field(None, alias="certificate")
    certificate_file: str = Field(None, alias="certificate_file")
    truststore_file: str = Field(None, alias="file_path")
    truststore_type: StoreType = Field(None, alias="store_type")
    truststore_password: str = Field(None, alias="password")
    use_expression_certificate: bool = Field(None, alias="is_expression_server_certificate")
    use_expression_certificate_file: bool = Field(None, alias="is_expression_server_certificate_file")
    use_expression_truststore_file: bool = Field(None, alias="is_expression_server_truststore_file")
    use_expression_truststore_password: bool = Field(None, alias="is_expression_server_truststore_password")

    @field_serializer("truststore_type")
    def serialize_truststore_type(self, truststore_type: StoreType | None) -> StoreType | None:
        if truststore_type:
            return truststore_type
        type = self.certificate_type.value if hasattr(self.certificate_type, "value") else self.certificate_type
        if type == "STORE_FILE":
            return StoreType.PKCS12
        return None

    @field_serializer("certificate")
    def serialize_certificate(self, certificate: str | dict | None) -> dict | None:
        if certificate is None:
            return certificate
        if isinstance(certificate, dict):
            return certificate
        if self.use_expression_certificate:
            return {"expression": certificate}
        return {"expression": f"`{certificate}`"}

    @field_serializer("certificate_file")
    def serialize_certificate_file(self, certificate_file: str | dict | None) -> dict | None:
        if certificate_file is None:
            return certificate_file
        if isinstance(certificate_file, dict):
            return certificate_file
        if self.use_expression_certificate_file:
            return {"expression": certificate_file}
        return {"expression": f"`{certificate_file}`"}

    @field_serializer("truststore_file")
    def serialize_truststore_file(self, truststore_file: str | dict | None) -> dict | None:
        if truststore_file is None:
            return truststore_file
        if isinstance(truststore_file, dict):
            return truststore_file
        if self.use_expression_truststore_file:
            return {"expression": truststore_file}
        return {"expression": f"`{truststore_file}`"}

    @field_serializer("truststore_password")
    def serialize_truststore_password(self, truststore_password: str | dict | None) -> dict | None:
        if truststore_password is None:
            return truststore_password
        if isinstance(truststore_password, dict):
            return truststore_password
        return {"expression": f"`{truststore_password}`"}


class ClientCertificate(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    certificate_type: CertificateType = Field(CertificateType.pem_text, alias="type")
    certificate: str = Field(None, alias="certificate")
    private_key: str = Field(None, alias="private_key")
    certificate_file: str = Field(None, alias="certificate_file")
    private_key_file: str = Field(None, alias="private_key_file")
    keystore_file: str = Field(None, alias="file_path")
    keystore_type: StoreType = Field(None, alias="store_type")
    keystore_password: str = Field(None, alias="password")
    use_expression_certificate: bool = Field(None, alias="is_expression_client_certificate")
    use_expression_private_key: bool = Field(None, alias="is_expression_client_private_key")
    use_expression_certificate_file: bool = Field(None, alias="is_expression_client_certificate_file")
    use_expression_private_key_file: bool = Field(None, alias="is_expression_client_private_key_file")
    use_expression_keystore_file: bool = Field(None, alias="is_expression_client_keystore_file")
    use_expression_keystore_password: bool = Field(None, alias="is_expression_client_keystore_password")

    @field_serializer("keystore_type")
    def serialize_keystore_type(self, keystore_type: StoreType | None) -> StoreType | None:
        if keystore_type:
            return keystore_type
        type = self.certificate_type.value if hasattr(self.certificate_type, "value") else self.certificate_type
        if type == "STORE_FILE":
            return StoreType.PKCS12
        return None

    @field_serializer("certificate")
    def serialize_certificate(self, certificate: str | dict | None) -> dict | None:
        if certificate is None:
            return certificate
        if isinstance(certificate, dict):
            return certificate
        elif self.use_expression_certificate:
            return {"expression": certificate}
        return {"expression": f"`{certificate}`"}

    @field_serializer("private_key")
    def serialize_private_key(self, private_key: str | dict | None) -> dict | None:
        if private_key is None:
            return private_key
        if isinstance(private_key, dict):
            return private_key
        elif self.use_expression_private_key:
            return {"expression": private_key}
        return {"expression": f"`{private_key}`"}

    @field_serializer("certificate_file")
    def serialize_certificate_file(self, certificate_file: str | dict | None) -> dict | None:
        if certificate_file is None:
            return certificate_file
        if isinstance(certificate_file, dict):
            return certificate_file
        elif self.use_expression_certificate_file:
            return {"expression": certificate_file}
        return {"expression": f"`{certificate_file}`"}

    @field_serializer("private_key_file")
    def serialize_private_key_file(self, private_key_file: str | dict | None) -> dict | None:
        if private_key_file is None:
            return private_key_file
        if isinstance(private_key_file, dict):
            return private_key_file
        elif self.use_expression_private_key_file:
            return {"expression": private_key_file}
        return {"expression": f"`{private_key_file}`"}

    @field_serializer("keystore_file")
    def serialize_keystore_file(self, keystore_file: str | dict | None) -> dict | None:
        if keystore_file is None:
            return keystore_file
        if isinstance(keystore_file, dict):
            return keystore_file
        elif self.use_expression_keystore_file:
            return {"expression": keystore_file}
        return {"expression": f"`{keystore_file}`"}

    @field_serializer("keystore_password")
    def serialize_keystore_password(self, keystore_password: str | dict | None) -> dict | None:
        if keystore_password is None:
            return keystore_password
        if isinstance(keystore_password, dict):
            return keystore_password
        elif self.use_expression_keystore_password:
            return {"expression": keystore_password}
        return {"expression": f"`{keystore_password}`"}


class Proxy(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    host: str = Field("", alias="host")
    port: str = Field("", alias="port")
    username: str = Field("", alias="username")
    password: str = Field("", alias="password")


class Settings(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    same_config: bool = Field(False, alias="inherit", description="The same configuration as request 0")
    enable_ssl: bool = Field(False, alias="enable_ssl")
    accept_self_signed_certificate: bool = Field(False, alias="accept_self_signed_certificate")
    verify_host_name: bool = Field(False, alias="verify_host")
    ssl_server_authentication: bool = Field(False, alias="enable_ssl_server")
    server_certificate: ServerCertificate = Field(None, alias="server_cert")
    ssl_client_authentication: bool = Field(False, alias="enable_ssl_client")
    client_certificate: ClientCertificate = Field(None, alias="client_cert")
    timeout_duration: int = Field(180000, alias="timeout_duration")
    retry_on_failure: int = Field(5, alias="retry_on_failure")
    use_proxy_server: bool = Field(False, alias="use_proxy")
    proxy: Proxy = Field(Proxy(), alias="proxy")
    action_on_failure: ActionOnFailure = Field(ActionOnFailure.abort, alias="on_error_control")

    @computed_field(alias="server_certificate")
    @property
    def server_cert(self) -> str | None:
        if self.server_certificate:
            return None
        return ""

    @computed_field(alias="client_certificate")
    @property
    def client_cert(self) -> str | None:
        if self.client_certificate:
            return None
        return ""


class Control(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    same_config: bool = Field(False, alias="inherit", description="The same configuration as request 0")
    enable_pagination: bool = Field(False, alias="iteration")
    pagination_condition: str = Field("", alias="prerequisite")
    data_output_control: DataOutputControl = Field(DataOutputControl.output_when_request_complete, alias="output_type")
    pre_action: str = Field("", alias="pre_action")
    post_action: str = Field("", alias="post_action")
    ignore_request_when_pre_action_false: bool = Field(False, alias="skip_on_pre_action_false")

    @field_serializer("pre_action")
    def serialize_pre_action(self, pre_action: str | dict | None) -> dict | None:
        if pre_action is None:
            return pre_action
        if isinstance(pre_action, dict):
            return pre_action
        return {"expression": pre_action}

    @field_serializer("post_action")
    def serialize_post_action(self, post_action: str | dict | None) -> dict | None:
        if post_action is None:
            return post_action
        if isinstance(post_action, dict):
            return post_action
        return {"expression": post_action}

    @field_serializer("pagination_condition")
    def serialize_(self, pagination_condition: str | dict | None) -> dict | None:
        if pagination_condition is None:
            return pagination_condition
        if isinstance(pagination_condition, dict):
            return pagination_condition
        return {"expression": pagination_condition}


class Request(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    method: Method = Field(Method.GET, alias="method")
    url: str = Field(None, alias="url")
    use_expression_url: bool = Field(None, alias="expression_endpoint_on")
    authentication: Authentication = Field(None, alias="authentication")
    request: RequestInfo = Field(None, alias="request")
    response: Response = Field(None, alias="response")
    settings: Settings = Field(None, alias="settings")
    control: Control = Field(None, alias="control")

    @computed_field
    @property
    def endpoint(self) -> dict:
        ep = {"method": self.method, "url": {"expression": f"`{self.url}`" or ""}}
        if self.use_expression_url:
            ep["expression_endpoint_on"] = self.use_expression_url
            ep["url"] = {"expression": self.url}
        return ep

    @model_serializer
    def serialize_model(self) -> dict[str, Any]:
        model_dict = {
            "endpoint": self.endpoint,
            "authentication": {
                "inherit": self.authentication.same_config,
                "body": self.authentication.model_dump(by_alias=True, exclude={"same_config"}, exclude_none=True),
            },
            "request": self.request.model_dump(
                exclude_none=True,
                by_alias=True,
                exclude={"body", "additional_headers_on"},
            ),
            "response": self.response.model_dump(exclude_none=True, by_alias=True, include={"same_config"}),
            "settings": self.settings.model_dump(exclude_none=True, by_alias=True),
            "control": self.control.model_dump(exclude_none=True, by_alias=True),
        }
        if self.request.body:
            model_dict["request"]["body"] = {
                "content": self.request.body.model_dump(
                    exclude_none=True,
                    by_alias=True,
                    exclude={
                        "form_data",
                        "form_urlencoded_data",
                        "file_path",
                        "raw_test",
                        "use_expression_text",
                        "use_expression_data",
                        "use_expression_file_path",
                        "expression_text_on",
                    },
                ),
                **self.request.body.model_dump(
                    exclude_none=True,
                    by_alias=True,
                    include={"use_expression_file_path", "expression_text_on"},
                ),
                **self.request.model_dump(
                    include={"additional_headers_on"},
                    exclude_none=True,
                    by_alias=True,
                    exclude_defaults=True,
                ),
            }
        model_dict["response"]["body"] = self.response.model_dump(
            exclude_none=True,
            by_alias=True,
            include={
                "content",
                "content_type",
                "encoding_type",
                "use_expression_file_path",
            },
        )

        if "params" in model_dict["request"]:
            for i, param in enumerate(model_dict["request"]["params"]):
                param["id"] = i
        if "headers" in model_dict["request"]:
            for i, header in enumerate(model_dict["request"]["headers"]):
                header["id"] = i
        if "cookies" in model_dict["request"]:
            for i, cookie in enumerate(model_dict["request"]["cookies"]):
                cookie["id"] = i
        if "body" in model_dict["request"] and "data" in model_dict["request"]["body"]["content"]:
            if isinstance(model_dict["request"]["body"]["content"]["data"], list):
                for i, data in enumerate(model_dict["request"]["body"]["content"]["data"]):
                    data["id"] = i
        return model_dict


class Variable(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    name: str = Field("", alias="name")
    data_type: DataType = Field("", alias="type")
    default_value: str = Field("", alias="default_value")
    derivation: str = Field("", alias="derivation")
    description: str = Field("", alias="description")

    @field_serializer("derivation")
    def serialize_(self, derivation: str | dict | None) -> dict | None:
        if derivation is None:
            return derivation
        if isinstance(derivation, dict):
            return derivation
        return {"expression": derivation}


class rest:
    Method = Method
    AuthenticationType = AuthenticationType
    AddTo = AddTo
    BodyType = BodyType
    CertificateType = CertificateType
    StoreType = StoreType
    ActionOnFailure = ActionOnFailure
    DataOutputControl = DataOutputControl
    DataType = DataType
    Authentication = Authentication
    Parameter = Parameter
    Header = Header
    Cookie = Cookie
    FormData = FormData
    FormURLEncodedData = FormURLEncodedData
    Body = Body
    RequestInfo = RequestInfo
    Response = Response
    ServerCertificate = ServerCertificate
    ClientCertificate = ClientCertificate
    Proxy = Proxy
    Settings = Settings
    Control = Control
    Request = Request
    Variable = Variable
